/*
 * Copyright 1999-2004 The Apache Software Foundation.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* $Id$ */
 
package org.apache.fop;

import java.io.File;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.sax.SAXResult;
import javax.xml.transform.stream.StreamSource;

import org.xml.sax.InputSource;

import org.apache.avalon.framework.container.ContainerUtil;
import org.apache.avalon.framework.logger.Logger;
import org.apache.avalon.framework.logger.NullLogger;
import org.apache.commons.io.output.ByteArrayOutputStream;
import org.apache.fop.apps.Driver;
import org.apache.fop.apps.InputHandler;
import org.apache.fop.apps.XSLTInputHandler;
import org.w3c.dom.Document;

/**
 * Basic runtime test for the old Driver class. It is used to verify that 
 * nothing obvious is broken after compiling.
 * @author <a href="mailto:jeremias@apache.org">Jeremias Maerki</a>
 */
public class BasicDriverTestCase extends AbstractFOPTestCase {

    private Logger logger = new NullLogger();

    /**
     * @see junit.framework.TestCase#TestCase(String)
     */
    public BasicDriverTestCase(String name) {
        super(name);
    }

    /**
     * Tests Driver with its special constructor for FO-->PDF conversion.
     * @throws Exception if anything fails
     */
    public void testFO2PDFWithConstructorSetup() throws Exception {
        File foFile = new File(getBaseDir(), "test/xml/bugtests/block.fo");
        ByteArrayOutputStream baout = new ByteArrayOutputStream();
        Driver driver = new Driver(
            new InputSource(foFile.toURL().toExternalForm()),
            baout);
        //Use deprecated method with purpose to validate backwards-compatibility.
        driver.setLogger(this.logger);
        driver.setRenderer(Driver.RENDER_PDF);
        driver.run();
        assertTrue("Generated PDF has zero length", baout.size() > 0);
    }

    /**
     * Tests Driver with InputSource and OutputStream.
     * @throws Exception if anything fails
     */
    public void testFO2PDFWithInputSource() throws Exception {
        File foFile = new File(getBaseDir(), "test/xml/bugtests/block.fo");
        ByteArrayOutputStream baout = new ByteArrayOutputStream();
        Driver driver = new Driver();
        //Use deprecated method with purpose to validate backwards-compatibility.
        driver.setLogger(this.logger);
        driver.setInputSource(new InputSource(foFile.toURL().toExternalForm()));
        driver.setOutputStream(baout);
        driver.setRenderer(Driver.RENDER_PDF);
        driver.run();
        assertTrue("Generated PDF has zero length", baout.size() > 0);
    }

    private Document loadDocument(File foFile) 
                throws TransformerException {
        TransformerFactory factory = TransformerFactory.newInstance();
        Transformer transformer = factory.newTransformer();
        Source src = new StreamSource(foFile);
        DOMResult res = new DOMResult();
        transformer.transform(src, res);
        return (Document)res.getNode();                                
    }

    /**
     * Tests Driver with Document and OutputStream.
     * @throws Exception if anything fails
     */
    public void testFO2PDFWithDOM() throws Exception {
        File foFile = new File(getBaseDir(), "test/xml/bugtests/block.fo");
        ByteArrayOutputStream baout = new ByteArrayOutputStream();
        Driver driver = new Driver();
        //Use deprecated method with purpose to validate backwards-compatibility.
        driver.setLogger(this.logger);
        driver.setOutputStream(baout);
        driver.setRenderer(Driver.RENDER_PDF);
        driver.render(loadDocument(foFile));
        assertTrue("Generated PDF has zero length", baout.size() > 0);
    }

    /**
     * Tests Driver with XMLReader, InputSource and OutputStream.
     * @throws Exception if anything fails
     */
    public void testFO2PDFWithXMLReader() throws Exception {
        File foFile = new File(getBaseDir(), "test/xml/bugtests/block.fo");
        ByteArrayOutputStream baout = new ByteArrayOutputStream();
        Driver driver = new Driver();
        //Use deprecated method with purpose to validate backwards-compatibility.
        driver.setLogger(this.logger);
        driver.setOutputStream(baout);
        driver.setRenderer(Driver.RENDER_PDF);
        SAXParserFactory factory = SAXParserFactory.newInstance();
        factory.setNamespaceAware(true);
        factory.setValidating(false);
        SAXParser parser = factory.newSAXParser();
        driver.render(parser.getXMLReader(),
            new InputSource(foFile.toURL().toExternalForm()));
        assertTrue("Generated PDF has zero length", baout.size() > 0);
    }

    /**
     * Tests Driver with JAXP and OutputStream generating PDF.
     * @throws Exception if anything fails
     */
    public void testFO2PDFWithJAXP() throws Exception {
        File foFile = new File(getBaseDir(), "test/xml/bugtests/block.fo");
        ByteArrayOutputStream baout = new ByteArrayOutputStream();
        Driver driver = new Driver();
        ContainerUtil.enableLogging(driver, this.logger);
        driver.setOutputStream(baout);
        driver.setRenderer(Driver.RENDER_PDF);
        
        TransformerFactory factory = TransformerFactory.newInstance();
        Transformer transformer = factory.newTransformer(); //Identity transf.
        Source src = new StreamSource(foFile);
        Result res = new SAXResult(driver.getContentHandler());
        transformer.transform(src, res);
        
        assertTrue("Generated PDF has zero length", baout.size() > 0);
    }

    /**
     * Tests Driver with JAXP and OutputStream generating PostScript.
     * @throws Exception if anything fails
     */
    public void testFO2PSWithJAXP() throws Exception {
        File foFile = new File(getBaseDir(), "test/xml/bugtests/block.fo");
        ByteArrayOutputStream baout = new ByteArrayOutputStream();
        Driver driver = new Driver();
        ContainerUtil.enableLogging(driver, this.logger);
        driver.setOutputStream(baout);
        driver.setRenderer(Driver.RENDER_PS);
        
        TransformerFactory factory = TransformerFactory.newInstance();
        Transformer transformer = factory.newTransformer(); //Identity transf.
        Source src = new StreamSource(foFile);
        Result res = new SAXResult(driver.getContentHandler());
        transformer.transform(src, res);
        
        assertTrue("Generated PostScript has zero length", baout.size() > 0);
    }

    /**
     * Tests Driver with JAXP and OutputStream generating RTF.
     * @throws Exception if anything fails
     */
    public void testFO2RTFWithJAXP() throws Exception {
        File foFile = new File(getBaseDir(), "test/xml/bugtests/block.fo");
        ByteArrayOutputStream baout = new ByteArrayOutputStream();
        Driver driver = new Driver();
        ContainerUtil.enableLogging(driver, this.logger);
        driver.setOutputStream(baout);
        driver.setRenderer(Driver.RENDER_RTF);
        
        TransformerFactory factory = TransformerFactory.newInstance();
        Transformer transformer = factory.newTransformer(); //Identity transf.
        Source src = new StreamSource(foFile);
        Result res = new SAXResult(driver.getContentHandler());
        transformer.transform(src, res);
        
        assertTrue("Generated RTF has zero length", baout.size() > 0);
    }

    /**
     * Tests Driver with XsltInputHandler and OutputStream.
     * @throws Exception if anything fails
     */
    public void testFO2PDFWithXSLTInputHandler() throws Exception {
        File xmlFile = new File(getBaseDir(), "test/xml/1.xml");
        File xsltFile = new File(getBaseDir(), "test/xsl/doc.xsl");
        ByteArrayOutputStream baout = new ByteArrayOutputStream();
        Driver driver = new Driver();
        ContainerUtil.enableLogging(driver, this.logger);
        driver.setOutputStream(baout);
        driver.setRenderer(Driver.RENDER_PDF);
        
        InputHandler handler = new XSLTInputHandler(xmlFile, xsltFile);
        driver.render(handler);
        
        assertTrue("Generated PDF has zero length", baout.size() > 0);
    }

}
